--- old/usr/src/uts/common/fs/zfs/dmu_send.c	2015-12-29 17:17:30.720430474 -0500
+++ new/usr/src/uts/common/fs/zfs/dmu_send.c	2015-12-29 17:17:30.726253997 -0500
@@ -24,6 +24,7 @@
  * Copyright (c) 2011, 2015 by Delphix. All rights reserved.
  * Copyright (c) 2014, Joyent, Inc. All rights reserved.
  * Copyright 2014 HybridCluster. All rights reserved.
+ * Copyright 2015 RackTop Systems.
  */
 
 #include <sys/dmu.h>
@@ -152,6 +153,13 @@
 	struct drr_free *drrf = &(dsp->dsa_drr->drr_u.drr_free);
 
 	/*
+	 * Skip free records if asked not to send them.  The resulting
+	 * stream cannot be received as a clone.
+	 */
+	if (dsp->dsa_skip_free)
+		return (0);
+
+	/*
 	 * When we receive a free record, dbuf_free_range() assumes
 	 * that the receiving system doesn't have any dbufs in the range
 	 * being freed.  This is always true because there is a one-record
@@ -346,6 +354,10 @@
 {
 	struct drr_freeobjects *drrfo = &(dsp->dsa_drr->drr_u.drr_freeobjects);
 
+	/* See comment in dump_free(). */
+	if (dsp->dsa_skip_free)
+		return (0);
+
 	/*
 	 * If there is a pending op, but it's not PENDING_FREEOBJECTS,
 	 * push it out, since free block aggregation can only be done for
@@ -684,7 +696,7 @@
 dmu_send_impl(void *tag, dsl_pool_t *dp, dsl_dataset_t *to_ds,
     zfs_bookmark_phys_t *ancestor_zb,
     boolean_t is_clone, boolean_t embedok, boolean_t large_block_ok, int outfd,
-    uint64_t resumeobj, uint64_t resumeoff,
+    uint64_t resumeobj, uint64_t resumeoff, boolean_t skip_free,
     vnode_t *vp, offset_t *off)
 {
 	objset_t *os;
@@ -745,7 +757,8 @@
 	drr->drr_u.drr_begin.drr_toguid = dsl_dataset_phys(to_ds)->ds_guid;
 	if (dsl_dataset_phys(to_ds)->ds_flags & DS_FLAG_CI_DATASET)
 		drr->drr_u.drr_begin.drr_flags |= DRR_FLAG_CI_DATA;
-	drr->drr_u.drr_begin.drr_flags |= DRR_FLAG_FREERECORDS;
+	if (!skip_free)
+		drr->drr_u.drr_begin.drr_flags |= DRR_FLAG_FREERECORDS;
 
 	if (ancestor_zb != NULL) {
 		drr->drr_u.drr_begin.drr_fromguid =
@@ -771,6 +784,7 @@
 	dsp->dsa_featureflags = featureflags;
 	dsp->dsa_resume_object = resumeobj;
 	dsp->dsa_resume_offset = resumeoff;
+	dsp->dsa_skip_free = skip_free;
 
 	mutex_enter(&to_ds->ds_sendstream_lock);
 	list_insert_head(&to_ds->ds_sendstreams, dsp);
@@ -874,7 +888,7 @@
 int
 dmu_send_obj(const char *pool, uint64_t tosnap, uint64_t fromsnap,
     boolean_t embedok, boolean_t large_block_ok,
-    int outfd, vnode_t *vp, offset_t *off)
+    int outfd, boolean_t skip_free, vnode_t *vp, offset_t *off)
 {
 	dsl_pool_t *dp;
 	dsl_dataset_t *ds;
@@ -910,10 +924,10 @@
 		is_clone = (fromds->ds_dir != ds->ds_dir);
 		dsl_dataset_rele(fromds, FTAG);
 		err = dmu_send_impl(FTAG, dp, ds, &zb, is_clone,
-		    embedok, large_block_ok, outfd, 0, 0, vp, off);
+		    embedok, large_block_ok, outfd, 0, 0, skip_free, vp, off);
 	} else {
 		err = dmu_send_impl(FTAG, dp, ds, NULL, B_FALSE,
-		    embedok, large_block_ok, outfd, 0, 0, vp, off);
+		    embedok, large_block_ok, outfd, 0, 0, skip_free, vp, off);
 	}
 	dsl_dataset_rele(ds, FTAG);
 	return (err);
@@ -922,7 +936,7 @@
 int
 dmu_send(const char *tosnap, const char *fromsnap, boolean_t embedok,
     boolean_t large_block_ok, int outfd, uint64_t resumeobj, uint64_t resumeoff,
-    vnode_t *vp, offset_t *off)
+    boolean_t skip_free, vnode_t *vp, offset_t *off)
 {
 	dsl_pool_t *dp;
 	dsl_dataset_t *ds;
@@ -990,11 +1004,11 @@
 		}
 		err = dmu_send_impl(FTAG, dp, ds, &zb, is_clone,
 		    embedok, large_block_ok,
-		    outfd, resumeobj, resumeoff, vp, off);
+		    outfd, resumeobj, resumeoff, skip_free, vp, off);
 	} else {
 		err = dmu_send_impl(FTAG, dp, ds, NULL, B_FALSE,
 		    embedok, large_block_ok,
-		    outfd, resumeobj, resumeoff, vp, off);
+		    outfd, resumeobj, resumeoff, skip_free, vp, off);
 	}
 	if (owned)
 		dsl_dataset_disown(ds, FTAG);
